using QEDprocesses
import QEDbase.mass

# TODO check
const e = sqrt(4π / 137)

"""
    QEDModel <: AbstractPhysicsModel

Singleton definition for identification of the QED-Model.
"""
struct QEDModel <: AbstractPhysicsModel end

"""
    QEDParticle

Base type for all particles in the [`QEDModel`](@ref).

Its template parameter specifies the particle's direction.

The concrete types contain singletons of the types that they are, like `Photon` and `Electron` from QEDbase, and their state descriptions.
"""
abstract type QEDParticle{Direction <: ParticleDirection} <: AbstractParticle end

"""
    QEDProcessDescription <: AbstractProcessDescription

A description of a process in the QED-Model. Contains the input and output particles.

See also: [`in_particles`](@ref), [`out_particles`](@ref), [`parse_process`](@ref)
"""
struct QEDProcessDescription <: AbstractProcessDescription
    inParticles::Dict{Type{<:QEDParticle{Incoming}}, Int}
    outParticles::Dict{Type{<:QEDParticle{Outgoing}}, Int}
end

"""
    QEDProcessInput <: AbstractProcessInput

Input for a QED Process. Contains the [`QEDProcessDescription`](@ref) of the process it is an input for, and the values of the in and out particles.

See also: [`gen_process_input`](@ref)
"""
struct QEDProcessInput <: AbstractProcessInput
    process::QEDProcessDescription
    inParticles::Vector{QEDParticle}
    outParticles::Vector{QEDParticle}
end

QEDParticleValue{ParticleType <: QEDParticle} = Union{
    ParticleValue{ParticleType, BiSpinor},
    ParticleValue{ParticleType, AdjointBiSpinor},
    ParticleValue{ParticleType, DiracMatrix},
    ParticleValue{ParticleType, SLorentzVector{Float64}},
    ParticleValue{ParticleType, ComplexF64},
}

"""
    PhotonStateful <: QEDParticle

A photon of the [`QEDModel`](@ref) with its state.
"""
struct PhotonStateful{Direction <: ParticleDirection} <: QEDParticle{Direction}
    momentum::SFourMomentum
    # this will maybe change to the full polarization vector? or do i need both
    polarization::AbstractDefinitePolarization
end

PhotonStateful{Direction}(mom::SFourMomentum) where {Direction <: ParticleDirection} =
    PhotonStateful{Direction}(mom, PolX()) # TODO: make allpol possible

PhotonStateful{Dir1}(ph::PhotonStateful{Dir2}) where {Dir1 <: ParticleDirection, Dir2 <: ParticleDirection} =
    PhotonStateful{Dir1}(ph.momentum, ph.polarization)

"""
    FermionStateful <: QEDParticle

A fermion of the [`QEDModel`](@ref) with its state.
"""
struct FermionStateful{Direction <: ParticleDirection} <: QEDParticle{Direction}
    momentum::SFourMomentum
    spin::AbstractDefiniteSpin
    # TODO: mass for electron/muon/tauon representation?
end

FermionStateful{Direction}(mom::SFourMomentum) where {Direction <: ParticleDirection} =
    FermionStateful{Direction}(mom, SpinUp()) # TODO: make allspin possible

FermionStateful{Dir1}(f::FermionStateful{Dir2}) where {Dir1 <: ParticleDirection, Dir2 <: ParticleDirection} =
    FermionStateful{Dir1}(f.momentum, f.spin)

"""
    AntiFermionStateful <: QEDParticle

An anti-fermion of the [`QEDModel`](@ref) with its state.
"""
struct AntiFermionStateful{Direction <: ParticleDirection} <: QEDParticle{Direction}
    momentum::SFourMomentum
    spin::AbstractDefiniteSpin
    # TODO: mass for electron/muon/tauon representation?
end

AntiFermionStateful{Direction}(mom::SFourMomentum) where {Direction <: ParticleDirection} =
    AntiFermionStateful{Direction}(mom, SpinUp()) # TODO: make allspin possible

AntiFermionStateful{Dir1}(f::AntiFermionStateful{Dir2}) where {Dir1 <: ParticleDirection, Dir2 <: ParticleDirection} =
    AntiFermionStateful{Dir1}(f.momentum, f.spin)

"""
    interaction_result(t1::Type{T1}, t2::Type{T2}) where {T1 <: QEDParticle, T2 <: QEDParticle}

For two given particle types that can interact, return the third.
"""
function interaction_result(t1::Type{T1}, t2::Type{T2}) where {T1 <: QEDParticle, T2 <: QEDParticle}
    @assert false "Invalid interaction between particles of types $t1 and $t2"
end

interaction_result(::Type{FermionStateful{Incoming}}, ::Type{FermionStateful{Outgoing}}) = PhotonStateful{Incoming}
interaction_result(::Type{FermionStateful{Incoming}}, ::Type{AntiFermionStateful{Incoming}}) = PhotonStateful{Incoming}
interaction_result(::Type{FermionStateful{Incoming}}, ::Type{<:PhotonStateful}) = FermionStateful{Outgoing}

interaction_result(::Type{FermionStateful{Outgoing}}, ::Type{FermionStateful{Incoming}}) = PhotonStateful{Incoming}
interaction_result(::Type{FermionStateful{Outgoing}}, ::Type{AntiFermionStateful{Outgoing}}) = PhotonStateful{Incoming}
interaction_result(::Type{FermionStateful{Outgoing}}, ::Type{<:PhotonStateful}) = FermionStateful{Incoming}

# antifermion mirror
interaction_result(::Type{AntiFermionStateful{Incoming}}, t2::Type{<:QEDParticle}) =
    interaction_result(FermionStateful{Outgoing}, t2)
interaction_result(::Type{AntiFermionStateful{Outgoing}}, t2::Type{<:QEDParticle}) =
    interaction_result(FermionStateful{Incoming}, t2)

# photon commutativity
interaction_result(t1::Type{<:PhotonStateful}, t2::Type{<:QEDParticle}) = interaction_result(t2, t1)

# but prevent stack overflow
function interaction_result(t1::Type{<:PhotonStateful}, t2::Type{<:PhotonStateful})
    @assert false "Invalid interaction between particles of types $t1 and $t2"
end

"""
    propagation_result(t1::Type{T}) where {T <: QEDParticle}

Return the type of the inverted direction. E.g.
"""
propagation_result(::Type{FermionStateful{Incoming}}) = FermionStateful{Outgoing}
propagation_result(::Type{FermionStateful{Outgoing}}) = FermionStateful{Incoming}
propagation_result(::Type{AntiFermionStateful{Incoming}}) = AntiFermionStateful{Outgoing}
propagation_result(::Type{AntiFermionStateful{Outgoing}}) = AntiFermionStateful{Incoming}
propagation_result(::Type{PhotonStateful{Incoming}}) = PhotonStateful{Outgoing}
propagation_result(::Type{PhotonStateful{Outgoing}}) = PhotonStateful{Incoming}

"""
    types(::QEDModel)

Return a Vector of the possible types of particle in the [`QEDModel`](@ref).
"""
function types(::QEDModel)
    return [
        PhotonStateful{Incoming},
        PhotonStateful{Outgoing},
        FermionStateful{Incoming},
        FermionStateful{Outgoing},
        AntiFermionStateful{Incoming},
        AntiFermionStateful{Outgoing},
    ]
end

# type piracy?
String(::Type{Incoming}) = "Incoming"
String(::Type{Outgoing}) = "Outgoing"

String(::Incoming) = "in"
String(::Outgoing) = "out"

function String(::Type{<:PhotonStateful})
    return "k"
end
function String(::Type{<:FermionStateful})
    return "e"
end
function String(::Type{<:AntiFermionStateful})
    return "p"
end

@inline particle(::PhotonStateful) = Photon()
@inline particle(::FermionStateful) = Electron()
@inline particle(::AntiFermionStateful) = Positron()

@inline momentum(p::PhotonStateful)::SFourMomentum = p.momentum
@inline momentum(p::FermionStateful)::SFourMomentum = p.momentum
@inline momentum(p::AntiFermionStateful)::SFourMomentum = p.momentum

@inline spin_or_pol(p::PhotonStateful)::AbstractPolarization = p.polarization
@inline spin_or_pol(p::FermionStateful)::AbstractSpin = p.spin
@inline spin_or_pol(p::AntiFermionStateful)::AbstractSpin = p.spin

@inline direction(::PhotonStateful{Dir}) where {Dir <: ParticleDirection} = Dir()
@inline direction(::FermionStateful{Dir}) where {Dir <: ParticleDirection} = Dir()
@inline direction(::AntiFermionStateful{Dir}) where {Dir <: ParticleDirection} = Dir()

@inline direction(::Type{PhotonStateful{Dir}}) where {Dir <: ParticleDirection} = Dir()
@inline direction(::Type{FermionStateful{Dir}}) where {Dir <: ParticleDirection} = Dir()
@inline direction(::Type{AntiFermionStateful{Dir}}) where {Dir <: ParticleDirection} = Dir()

@inline isincoming(::QEDParticle{Incoming}) = true
@inline isincoming(::QEDParticle{Outgoing}) = false
@inline isoutgoing(::QEDParticle{Incoming}) = false
@inline isoutgoing(::QEDParticle{Outgoing}) = true

@inline isincoming(::Type{<:QEDParticle{Incoming}}) = true
@inline isincoming(::Type{<:QEDParticle{Outgoing}}) = false
@inline isoutgoing(::Type{<:QEDParticle{Incoming}}) = false
@inline isoutgoing(::Type{<:QEDParticle{Outgoing}}) = true

@inline mass(::Type{<:FermionStateful}) = 1.0
@inline mass(::Type{<:AntiFermionStateful}) = 1.0
@inline mass(::Type{<:PhotonStateful}) = 0.0

@inline invert_momentum(p::FermionStateful{Dir}) where {Dir <: ParticleDirection} =
    FermionStateful{Dir}(-p.momentum, p.spin)
@inline invert_momentum(p::AntiFermionStateful{Dir}) where {Dir <: ParticleDirection} =
    AntiFermionStateful{Dir}(-p.momentum, p.spin)
@inline invert_momentum(k::PhotonStateful{Dir}) where {Dir <: ParticleDirection} =
    PhotonStateful{Dir}(-k.momentum, k.polarization)


"""
    caninteract(T1::Type{<:QEDParticle}, T2::Type{<:QEDParticle})

For two given [`QEDParticle`](@ref) types, return whether they can interact at a vertex. This is equivalent to `!issame(T1, T2)`.

See also: [`issame`](@ref) and [`interaction_result`](@ref)
"""
function caninteract(T1::Type{<:QEDParticle}, T2::Type{<:QEDParticle})
    if (T1 == T2)
        return false
    end
    if (T1 <: PhotonStateful && T2 <: PhotonStateful)
        return false
    end

    for (P1, P2) in [(T1, T2), (T2, T1)]
        if (P1 == FermionStateful{Incoming} && P2 == AntiFermionStateful{Outgoing})
            return false
        end
        if (P1 == FermionStateful{Outgoing} && P2 == AntiFermionStateful{Incoming})
            return false
        end
    end

    return true
end

function type_index_from_name(::QEDModel, name::String)
    if startswith(name, "ki")
        return (PhotonStateful{Incoming}, parse(Int, name[3:end]))
    elseif startswith(name, "ko")
        return (PhotonStateful{Outgoing}, parse(Int, name[3:end]))
    elseif startswith(name, "ei")
        return (FermionStateful{Incoming}, parse(Int, name[3:end]))
    elseif startswith(name, "eo")
        return (FermionStateful{Outgoing}, parse(Int, name[3:end]))
    elseif startswith(name, "pi")
        return (AntiFermionStateful{Incoming}, parse(Int, name[3:end]))
    elseif startswith(name, "po")
        return (AntiFermionStateful{Outgoing}, parse(Int, name[3:end]))
    else
        throw("Invalid name for a particle in the QED model")
    end
end

"""
    issame(T1::Type{<:QEDParticle}, T2::Type{<:QEDParticle})

For two given [`QEDParticle`](@ref) types, return whether they are equivalent for the purpose of a Feynman Diagram. That means e.g. an `Incoming` `AntiFermion` is the same as an `Outgoing` `Fermion`. This is equivalent to `!caninteract(T1, T2)`.

See also: [`caninteract`](@ref) and [`interaction_result`](@ref)
"""
function issame(T1::Type{<:QEDParticle}, T2::Type{<:QEDParticle})
    return !caninteract(T1, T2)
end

"""
    QED_vertex()

Return the factor of a vertex in a QED feynman diagram.
"""
@inline function QED_vertex()::SLorentzVector{DiracMatrix}
    # Peskin-Schroeder notation
    return -1im * e * gamma()
end

@inline function QED_inner_edge(p::QEDParticle)
    pos_mom = p.momentum
    return propagator(particle(p), pos_mom)
end

"""
    QED_conserve_momentum(p1::QEDParticle, p2::QEDParticle)

Calculate and return a new particle from two given interacting ones at a vertex.
"""
function QED_conserve_momentum(p1::QEDParticle, p2::QEDParticle)
    #println("Conserving momentum of \n$(direction(p1)) $(p1)\n and \n$(direction(p2)) $(p2)")
    T3 = interaction_result(typeof(p1), typeof(p2))
    # TODO: probably also need to do something about the spin/pol
    p1_mom = p1.momentum
    if (typeof(direction(p1)) <: Outgoing)
        p1_mom *= -1
    end
    p2_mom = p2.momentum
    if (typeof(direction(p2)) <: Outgoing)
        p2_mom *= -1
    end

    p3_mom = p1_mom + p2_mom
    if (typeof(direction(T3)) <: Incoming)
        return T3(-p3_mom)
    end
    return T3(p3_mom)
end

"""
    model(::AbstractProcessDescription)

Return the model of this process description.
"""
model(::QEDProcessDescription) = QEDModel()
model(::QEDProcessInput) = QEDModel()

==(p1::QEDProcessDescription, p2::QEDProcessDescription) =
    p1.inParticles == p2.inParticles && p1.outParticles == p2.outParticles

function in_particles(process::QEDProcessDescription)
    return process.inParticles
end

function in_particles(input::QEDProcessInput)
    return input.inParticles
end

function out_particles(process::QEDProcessDescription)
    return process.outParticles
end

function out_particles(input::QEDProcessInput)
    return input.outParticles
end
